#include "FontRenderer.hpp"
#include "BitStream.hpp"

#include <iostream>
#include <algorithm>
#include <stdexcept>
#include <cassert>

#define FT_RENDER_FLAG (FT_LOAD_RENDER | FT_LOAD_MONOCHROME | FT_LOAD_TARGET_MONO)

FontRenderer::FontRenderer()
  : m_face(nullptr)
{
  if(FT_Init_FreeType(&m_ft))
    throw std::runtime_error("Unable to initialize FreeType");
}

FontRenderer::~FontRenderer()
{
  FT_Done_FreeType(m_ft);
}

void FontRenderer::LoadFont(const std::string& filename, int size)
{
  // Free old font
  if(m_face)
    FT_Done_Face(m_face);

  // Load new font
  if(FT_New_Face(m_ft, filename.c_str(), 0, &m_face))
    throw std::runtime_error("Unable to open font");
  FT_Set_Pixel_Sizes(m_face, 0, size);

  // Calculate font metrics
  m_metrics.ascender = m_metrics.descender = m_metrics.width = 0;
  for(int i = 0x20; i < 0x80; ++i)
  {
    // Render in monochrome
    if(FT_Load_Char(m_face, i, FT_RENDER_FLAG))
      continue;
    FT_GlyphSlot g = m_face->glyph;

    // Update metrics
    if(g->bitmap_top > m_metrics.ascender)
      m_metrics.ascender = g->bitmap_top;

    int descender = g->bitmap.rows - g->bitmap_top;
    if(descender > m_metrics.descender)
      m_metrics.descender = descender;

    int width = g->bitmap.width;
    if(width > m_metrics.width)
      m_metrics.width = width;
  }
}

Glyph FontRenderer::RenderGlyph(int character)
{
  // Render in monochrome
  if(FT_Load_Char(m_face, character, FT_RENDER_FLAG))
    return Glyph();
  FT_GlyphSlot g = m_face->glyph;

  // Check font
  if(g->advance.y != 0)
    std::cerr << "Error: Vertical fonts are not supported!" << std::endl;

  // Create glyph
  Glyph gly;
  gly.width = g->bitmap.width;
  gly.height = g->bitmap.rows;
  gly.xPos = g->bitmap_left;
  gly.yPos = m_metrics.ascender - g->bitmap_top;
  gly.advance = g->advance.x / 64;

  // Fill buffer
  BitStream bs;
  for(int row = 0; row < g->bitmap.rows; ++row)
  {
    for(int i = 0; 8*i < g->bitmap.width; ++i)
      bs.putBits(g->bitmap.buffer[row*g->bitmap.pitch + i], std::min(8, g->bitmap.width - 8*i));
  }
  gly.buffer = bs.getData();

  return gly;
}